/*\

Copyright 2017 The University of North Carolina at Chapel Hill.
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation for educational, research and non-profit purposes, without
fee, and without a written agreement is hereby granted, provided that the
above copyright notice and the following three paragraphs appear in all
copies.

IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
DAMAGES.

THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The authors may be contacted via:

US Mail: Real-Time Systems Group at UNC
Department of Computer Science
Sitterson Hall
University of N. Carolina
Chapel Hill, NC 27599-3175

EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu

**/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <assert.h>
#include <sched.h>
#include <inttypes.h>
#ifdef __APPLE__
#include <sys/uio.h>
#else
#include <sys/io.h>
#endif
#include <sys/mman.h>
#include <time.h>

#include "cycles.h"
#include "nesting.h"
#include "mem.h"
#include "fast_rwrnlp.h"

#define GHZ 2297.411
#define CONVERT_CYCLE_MSEC 2297.411

int nr_cpus;
int nr_requested;
int nr_resources;
float cs_length;
int nestedprob = 0;
int readprob = 0;

fast_rwrnlp_t fast_rwrnlp;

int shared_objects[sizeof(resource_mask_t)*20];

static volatile int counter = 0;

struct thread_context {
	int id;
	int* objects;
};

#ifndef __APPLE__
int migrate_thread_to_cpu (int target_cpu)
{
	cpu_set_t *cpu_set;
	size_t sz;
	int ret;

	cpu_set = CPU_ALLOC(nr_cpus);
	sz = CPU_ALLOC_SIZE(nr_cpus);
	CPU_ZERO_S(sz, cpu_set);
	CPU_SET_S(target_cpu, sz, cpu_set);

	ret = sched_setaffinity(0, sz, cpu_set);

	CPU_FREE(cpu_set);

	return ret;
}
#endif

resource_mask_t random_resources (int requested, int available)
{
    int i, j, tmp;
    bool flag;
    int choices[nr_resources];
    resource_mask_t retval = 0;
    if (available == 1) {
        tmp = rand() % available;
        retval |= (1l << tmp);
        return retval;
    }
    else if (available == requested) {
        for (i = 0; i < requested; i++)
        {
            retval |= (1l << i);
        }
    }
    else
    {
        for (i = 0; i < requested; i++)
        {
            do
            {
                tmp = (rand() % (available));// - 1));
                flag = true;

                for (j = 0; j < i; j++)
                {
                    if (tmp == choices[j])
                        flag = false;
                }
            } while (!flag);

            choices[i] = tmp;
            retval |= (1l << (tmp));// + 1l));
        }
    }

    return retval;
}

void* rt_thread (void* args)
{
	int i, j;
	struct thread_context* ctx = (struct thread_context*) args;
	uint64_t end_time;
	unsigned int seed = time(NULL)+ctx->id;

	int request_nonnested;
	resource_mask_t request_nested;
	bool isread;
	bool isnested;

#ifndef __APPLE__
	migrate_thread_to_cpu(ctx->id);
#endif

	// Randomly choose read or write
	if (rand_r(&seed) % 100 < readprob) {
		isread = true;
	} else {
		isread = false;
	}

	// Randomly choose nested or non-nested
	if (rand_r(&seed) % 100 < nestedprob) {
		isnested = true;
	}
	else {
		isnested = false;
	}

	// Randomly choose resources
	if (isnested)
	{
		request_nested = random_resources(nr_requested, nr_resources);
	}
	else
	{
		request_nested = random_resources(1, nr_resources);
		request_nonnested = ffsl(request_nested) - 1; // 1-indexed
	}

	// Ensure all tasks are in-sync
	__sync_fetch_and_add(&counter, 1);
	while (counter != nr_cpus) {
		cpu_relax();
	}
	mb();

	// Lock using the Fast RW-RNLP
	if (isnested)
	{
		if (isread)
		{
			n_read_lock(&fast_rwrnlp, request_nested, ctx->id);
		}
		else
		{
			n_write_lock(&fast_rwrnlp, request_nested, ctx->id);
		}
	}
	else
	{
		if (isread) {
			nn_read_lock(&fast_rwrnlp, request_nonnested);
		}
		else
		{
			nn_write_lock(&fast_rwrnlp, request_nonnested);
		}
	}

	/*******code for critical section*******/
	end_time = get_cycles() + (cs_length * CONVERT_CYCLE_MSEC); // * 1000 should be 1ms
	while (get_cycles() < end_time) {
		cpu_relax();
	}
	/****************************************/
	__sync_synchronize();

	// Unlock the Fast RW-RNLP
	if (isnested)
	{
		if (isread)
		{
			n_read_unlock(&fast_rwrnlp, request_nested, ctx->id);
		}
		else
		{
			n_write_unlock(&fast_rwrnlp, request_nested, ctx->id);
		}
	}
	else
	{
		if (isread)
		{
			nn_read_unlock(&fast_rwrnlp, request_nonnested);
		}
		else
		{
			nn_write_unlock(&fast_rwrnlp, request_nonnested);
		}
	}

	printf("Success!\n");

	return NULL;
}

static void usage(char *error)
{
	fprintf(stderr, "Error: %s\n", error);
	fprintf(stderr,
		"Usage:\n"
		"	lprun	[-n CPU-NUMBER]\n"
		"		[-q NR_RESOURCES] [-k NR_REQUESTED]\n"
		"		[-c CS_LENGTH]\n"
		"		[-p nested probability] [-P read probability]\n"
		"\n"
		"CRITICAL SECTION LENGTH is in microseconds.\n");
	exit(EXIT_FAILURE);
}

#define OPTSTR "n:q:k:c:p:P:"
int main(int argc, char** argv)
{
	int i, count;
	pthread_t *tasks;
	struct thread_context* ctx;

	mlockall(MCL_CURRENT | MCL_FUTURE);

	count = 0;
	int opt;
	while ((opt = getopt(argc, argv, OPTSTR)) != -1) {
		switch (opt) {
		case 'n':
			nr_cpus = atoi(optarg);
			count++;
			break;
		case 'q':
			nr_resources = atoi(optarg);
			count++;
			break;
		case 'k':
			nr_requested = atoi(optarg);
			count++;
			break;
		case 'c':
			cs_length = atoi(optarg);
			count++;
			break;
		case 'p':
			nestedprob = atoi(optarg);
			count++;
			break;
		case 'P':
			readprob = atoi(optarg);
			count++;
			break;
		case ':':
			usage("Argument missing.");
			exit(EXIT_FAILURE);
		case '?':
		default:
			usage("Bad argument.");
			exit(EXIT_FAILURE);
		}
	}

	if (count < 6) {
		usage("Missing required options");
		exit(EXIT_FAILURE);
	}

	tasks = calloc(nr_cpus, sizeof(pthread_t));
	ctx = malloc(nr_cpus * sizeof(struct thread_context));

	fast_rwrnlp_init(&fast_rwrnlp);

	for(i = 0; i < nr_cpus; i++)
	{
		ctx[i].id = i;
		ctx[i].objects = shared_objects;
	}

	for(i = 0; i < nr_cpus; i++)
	{
		mb();
		pthread_create(&tasks[i], NULL, rt_thread, (void *) (&ctx[i]));
		mb();
	}

	for(i = 0; i < nr_cpus; i++)
	{
		pthread_join(tasks[i], NULL);
	}

	return 0;
}
